@@ -6,43 +6,44 @@ This module defines the [`Download`] type and support code.
 !*/
 
 use crate::macros::*;
-use crate::{value_conversion, Error, File, Peer, Result, Server, Tracker};
+use crate::{value_conversion, File, Peer, Result, Server, Tracker};
+use crate::multicall::MultiBuilder;
 use std::sync::Arc;
 use xmlrpc::{Request, Value};
 
 macro_rules! d_getter {
-    ($(#[$meta:meta])* $method: ident, $result: ty, $conv: ident) => {
-        prim_getter!($(#[$meta])* "d.", $method, $result, $conv);
+    ($(#[$meta:meta])* $method: ident, $result: ty) => {
+        prim_getter!($(#[$meta])* "d.", $method, $result);
     }
 }
 
 macro_rules! d_str_getter {
     ($(#[$meta:meta])* $method: ident) => {
-        d_getter!($(#[$meta])* $method, String, string_owned);
+        d_getter!($(#[$meta])* $method, String);
     }
 }
 
 macro_rules! d_bool_getter {
     ($(#[$meta:meta])* $method: ident) => {
-        d_getter!($(#[$meta])* $method, bool, bool);
+        d_getter!($(#[$meta])* $method, bool);
     }
 }
 
 macro_rules! d_f1000_getter {
     ($(#[$meta:meta])* $method: ident) => {
-        d_getter!($(#[$meta])* $method, f64, fraction1000);
+        d_getter!($(#[$meta])* $method, f64);
     }
 }
 
 macro_rules! d_int_getter {
     ($(#[$meta:meta])* $method: ident) => {
-        d_getter!($(#[$meta])* $method, i64, int);
+        d_getter!($(#[$meta])* $method, i64);
     }
 }
 
 macro_rules! d_int_getter_named {
     ($(#[$meta:meta])* $method: ident, $apimethod: literal) => {
-        prim_getter_named!($(#[$meta])* "d.", $method, i64, int, $apimethod);
+        prim_getter_named!($(#[$meta])* "d.", $method, i64, $apimethod);
     }
 }
 
@@ -139,26 +140,22 @@ impl Download {
         &self.inner.sha1_hex
     }
 
+    /// Perform a "p.multicall", with a single accessor.
+    fn pmulticall1<T: TryFromValue>(&self, get1: &str) -> Result<Vec<(T,)>> {
+        let builder = MultiBuilder::new(
+            &self.inner.server,
+            "p.multicall",
+            self.sha1_hex(),
+            "");
+        builder.call(&format!("p.{}", get1)).invoke()
+    }
+
     /// Get a list of active peers associated with this download.
     pub fn peers(&self) -> Result<Vec<Peer>> {
-        let raw_list = Request::new("p.multicall")
-            .arg(self.sha1_hex())
-            .arg("")
-            .arg(Value::Array(vec!["p.id=".into()]))
-            .call_url(self.endpoint())?;
-        let list = value_conversion::list(&raw_list)?
-            .iter()
-            .map(|ll| {
-                let peerhash = value_conversion::list(ll)?
-                    .get(0)
-                    .ok_or(Error::UnexpectedStructure(
-                            format!("expected non-empty inner list, got {:?}", ll)
-                        ))?;
-                let peerhash = value_conversion::string(peerhash)?;
-                Ok(Peer::new(self.clone(), peerhash))
-            })
-            .collect();
-        list
+        self.pmulticall1::<String>("id")?
+            .into_iter()
+            .map(|(id,)| Ok(Peer::new(self.clone(), &id)))
+            .collect()
     }
 
     /// Get a list of files associated with this download.
@@ -197,6 +194,21 @@ impl Download {
         /// The number of completed chunks (pieces).
         completed_chunks);
 
+    d_int_getter_named!(
+        /// Get the download rate.
+        down_rate, "down.rate");
+    d_int_getter_named!(
+        /// Get the download total (bytes).
+        down_total, "down.total");
+
+    d_bool_getter!(is_active);
+    d_bool_getter!(is_open);
+    d_bool_getter!(is_closed);
+
+    d_str_getter!(
+        /// The metafile from which this download was created.
+        loaded_file);
+
     d_str_getter!(
         /// Unstructured error messages, either generated by rtorrent, or forwarded from the
         /// tracker.
@@ -205,46 +217,35 @@ impl Download {
     d_str_getter!(
         /// Get the name of the torrent.
         name);
+
+    d_f1000_getter!(
+        /// Get the upload/download ratio for this download.
+        ratio);
+
     d_int_getter!(
         /// Get the size, in bytes, of the torrent contents.
         size_bytes);
-    d_str_getter!(
-        /// The metafile from which this download was created.
-        loaded_file);
+    d_int_getter!(
+        /// Get the number of files associated with this download.
+        size_files);
+    d_bool_getter!(
+        /// Get the state (`false` is stopped).
+        state);
+
     d_str_getter!(
         /// Starts as the file the download was initially created from.
         tied_to_file);
 
-    d_bool_getter!(
-        /// Get the state (`false` is stopped).
-        state);
-    d_bool_getter!(is_active);
-    d_bool_getter!(is_open);
-    d_bool_getter!(is_closed);
-
     d_int_getter!(
         /// Get the number of trackers associated with this download.
         tracker_size);
-    d_int_getter!(
-        /// Get the number of files associated with this download.
-        size_files);
 
-    d_int_getter_named!(
-        /// Get the download rate.
-        down_rate, "down.rate");
-    d_int_getter_named!(
-        /// Get the download total (bytes).
-        down_total, "down.total");
     d_int_getter_named!(
         /// Get the upload rate.
         up_rate, "up.rate");
     d_int_getter_named!(
         /// Get the upload total (bytes).
         up_total, "up.total");
-
-    d_f1000_getter!(
-        /// Get the upload/download ratio for this download.
-        ratio);
 }
 
 unsafe impl Send for Download {}
